# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for getting BIP44 coins configuration."""

# Imports
from typing import Dict

from bip_utils.bip.conf.bip44.bip44_coins import Bip44Coins
from bip_utils.bip.conf.bip44.bip44_conf import Bip44Conf
from  import BipCoinConf, BipCoins


class Bip44ConfGetterConst:
    """Class container for BIP44 configuration getter constants."""

    # Map from Bip44Coins to configuration classes
    COIN_TO_CONF: Dict[BipCoins, BipCoinConf] = {
        Bip44Coins.AKASH_NETWORK: Bip44Conf.AkashNetwork,
        Bip44Coins.ALGORAND: Bip44Conf.Algorand,
        Bip44Coins.APTOS: Bip44Conf.Aptos,
        Bip44Coins.ARBITRUM: Bip44Conf.Arbitrum,
        Bip44Coins.AVAX_C_CHAIN: Bip44Conf.AvaxCChain,
        Bip44Coins.AVAX_P_CHAIN: Bip44Conf.AvaxPChain,
        Bip44Coins.AVAX_X_CHAIN: Bip44Conf.AvaxXChain,
        Bip44Coins.AXELAR: Bip44Conf.Axelar,
        Bip44Coins.BAND_PROTOCOL: Bip44Conf.BandProtocol,
        Bip44Coins.BINANCE_CHAIN: Bip44Conf.BinanceChain,
        Bip44Coins.BINANCE_SMART_CHAIN: Bip44Conf.BinanceSmartChain,
        Bip44Coins.BITCOIN: Bip44Conf.BitcoinMainNet,
        Bip44Coins.BITCOIN_REGTEST: Bip44Conf.BitcoinRegTest,
        Bip44Coins.BITCOIN_TESTNET: Bip44Conf.BitcoinTestNet,
        Bip44Coins.BITCOIN_CASH: Bip44Conf.BitcoinCashMainNet,
        Bip44Coins.BITCOIN_CASH_TESTNET: Bip44Conf.BitcoinCashTestNet,
        Bip44Coins.BITCOIN_CASH_SLP: Bip44Conf.BitcoinCashSlpMainNet,
        Bip44Coins.BITCOIN_CASH_SLP_TESTNET: Bip44Conf.BitcoinCashSlpTestNet,
        Bip44Coins.BITCOIN_SV: Bip44Conf.BitcoinSvMainNet,
        Bip44Coins.BITCOIN_SV_TESTNET: Bip44Conf.BitcoinSvTestNet,
        Bip44Coins.CARDANO_BYRON_ICARUS: Bip44Conf.CardanoByronIcarus,
        Bip44Coins.CARDANO_BYRON_LEDGER: Bip44Conf.CardanoByronLedger,
        Bip44Coins.CELESTIA: Bip44Conf.Celestia,
        Bip44Coins.CELO: Bip44Conf.Celo,
        Bip44Coins.CERTIK: Bip44Conf.Certik,
        Bip44Coins.CHIHUAHUA: Bip44Conf.Chihuahua,
        Bip44Coins.COSMOS: Bip44Conf.Cosmos,
        Bip44Coins.DASH: Bip44Conf.DashMainNet,
        Bip44Coins.DASH_TESTNET: Bip44Conf.DashTestNet,
        Bip44Coins.DOGECOIN: Bip44Conf.DogecoinMainNet,
        Bip44Coins.DOGECOIN_TESTNET: Bip44Conf.DogecoinTestNet,
        Bip44Coins.DYDX: Bip44Conf.DYDX,
        Bip44Coins.ECASH: Bip44Conf.EcashMainNet,
        Bip44Coins.ECASH_TESTNET: Bip44Conf.EcashTestNet,
        Bip44Coins.ELROND: Bip44Conf.Elrond,
        Bip44Coins.EOS: Bip44Conf.Eos,
        Bip44Coins.ERGO: Bip44Conf.ErgoMainNet,
        Bip44Coins.ERGO_TESTNET: Bip44Conf.ErgoTestNet,
        Bip44Coins.ETHEREUM: Bip44Conf.Ethereum,
        Bip44Coins.ETHEREUM_CLASSIC: Bip44Conf.EthereumClassic,
        Bip44Coins.FANTOM_OPERA: Bip44Conf.FantomOpera,
        Bip44Coins.FETCH_AI: Bip44Conf.FetchAi,
        Bip44Coins.FETCH_AI_ETH: Bip44Conf.FetchAiEth,
        Bip44Coins.FILECOIN: Bip44Conf.Filecoin,
        Bip44Coins.HARMONY_ONE_ATOM: Bip44Conf.HarmonyOneAtom,
        Bip44Coins.HARMONY_ONE_ETH: Bip44Conf.HarmonyOneEth,
        Bip44Coins.HARMONY_ONE_METAMASK: Bip44Conf.HarmonyOneMetamask,
        Bip44Coins.HUOBI_CHAIN: Bip44Conf.HuobiChain,
        Bip44Coins.ICON: Bip44Conf.Icon,
        Bip44Coins.INJECTIVE: Bip44Conf.Injective,
        Bip44Coins.IRIS_NET: Bip44Conf.IrisNet,
        Bip44Coins.KAVA: Bip44Conf.Kava,
        Bip44Coins.KUSAMA_ED25519_SLIP: Bip44Conf.KusamaEd25519Slip,
        Bip44Coins.LITECOIN: Bip44Conf.LitecoinMainNet,
        Bip44Coins.LITECOIN_TESTNET: Bip44Conf.LitecoinTestNet,
        Bip44Coins.METIS: Bip44Conf.Metis,
        Bip44Coins.MONERO_ED25519_SLIP: Bip44Conf.MoneroEd25519Slip,
        Bip44Coins.MONERO_SECP256K1: Bip44Conf.MoneroSecp256k1,
        Bip44Coins.MULTIVERSX: Bip44Conf.Elrond,
        Bip44Coins.NANO: Bip44Conf.Nano,
        Bip44Coins.NEAR_PROTOCOL: Bip44Conf.NearProtocol,
        Bip44Coins.NEO: Bip44Conf.NeoLegacy,
        Bip44Coins.NEO_LEGACY: Bip44Conf.NeoLegacy,
        Bip44Coins.NEO_N3: Bip44Conf.NeoN3,
        Bip44Coins.NEUTRON: Bip44Conf.Neutron,
        Bip44Coins.NIMIQ: Bip44Conf.Nimiq,
        Bip44Coins.NINE_CHRONICLES_GOLD: Bip44Conf.NineChroniclesGold,
        Bip44Coins.OKEX_CHAIN_ATOM: Bip44Conf.OkexChainAtom,
        Bip44Coins.OKEX_CHAIN_ATOM_OLD: Bip44Conf.OkexChainAtomOld,
        Bip44Coins.OKEX_CHAIN_ETH: Bip44Conf.OkexChainEth,
        Bip44Coins.ONTOLOGY: Bip44Conf.Ontology,
        Bip44Coins.OPTIMISM: Bip44Conf.Optimism,
        Bip44Coins.OSMOSIS: Bip44Conf.Osmosis,
        Bip44Coins.PI_NETWORK: Bip44Conf.PiNetwork,
        Bip44Coins.POLKADOT_ED25519_SLIP: Bip44Conf.PolkadotEd25519Slip,
        Bip44Coins.POLYGON: Bip44Conf.Polygon,
        Bip44Coins.RIPPLE: Bip44Conf.Ripple,
        Bip44Coins.SECRET_NETWORK_OLD: Bip44Conf.SecretNetworkOld,
        Bip44Coins.SECRET_NETWORK_NEW: Bip44Conf.SecretNetworkNew,
        Bip44Coins.SOLANA: Bip44Conf.Solana,
        Bip44Coins.STAFI: Bip44Conf.Stafi,
        Bip44Coins.STELLAR: Bip44Conf.Stellar,
        Bip44Coins.SUI: Bip44Conf.Sui,
        Bip44Coins.TERRA: Bip44Conf.Terra,
        Bip44Coins.TEZOS: Bip44Conf.Tezos,
        Bip44Coins.THETA: Bip44Conf.Theta,
        Bip44Coins.TRON: Bip44Conf.Tron,
        Bip44Coins.VECHAIN: Bip44Conf.VeChain,
        Bip44Coins.VERGE: Bip44Conf.Verge,
        Bip44Coins.ZCASH: Bip44Conf.ZcashMainNet,
        Bip44Coins.ZCASH_TESTNET: Bip44Conf.ZcashTestNet,
        Bip44Coins.ZILLIQA: Bip44Conf.Zilliqa,
    }


class Bip44ConfGetter:
    """
    BIP44 configuration getter class.
    It allows to get the BIP44 configuration of a specific coin.
    """

    @staticmethod
    def GetConfig(coin_type: BipCoins) -> BipCoinConf:
        """
        Get coin configuration.

        Args:
            coin_type (BipCoins): Coin type

        Returns:
            BipCoinConf: Coin configuration

        Raises:
            TypeError: If coin type is not of a Bip44Coins enumerative
        """
        if not isinstance(coin_type, Bip44Coins):
            raise TypeError("Coin type is not an enumerative of Bip44Coins")
        return Bip44ConfGetterConst.COIN_TO_CONF[coin_type]



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for BIP32 extended key serialization/deserialization."""

# Imports
from typing import Tuple

from ..base58 import Base58Decoder, Base58Encoder
from .bip32_const import Bip32Const
from .bip32_ex import Bip32KeyError
from .bip32_key_data import Bip32ChainCode, Bip32Depth, Bip32FingerPrint, Bip32KeyData, Bip32KeyIndex
from .bip32_key_net_ver import Bip32KeyNetVersions
from ..ecc import IPrivateKey, IPublicKey
from ..utils.misc import BytesUtils


class Bip32KeySerConst:
    """Class container for BIP32 key serialize constants."""

    # Serialized public key length in bytes
    SERIALIZED_PUB_KEY_BYTE_LEN: int = 78
    # Serialized private key length in bytes
    SERIALIZED_PRIV_KEY_BYTE_LEN: Tuple[int, int] = (78, 110)


class _Bip32KeySerializer:
    """
    BIP32 key serializer class.
    It serializes private/public keys.
    """

    @staticmethod
    def Serialize(key_bytes: bytes,
                  key_data: Bip32KeyData,
                  key_net_ver_bytes: bytes) -> str:
        """
        Serialize the specified key bytes.

        Args:
            key_bytes (bytes)           : Key bytes
            key_data (BipKeyData object): Key data
            key_net_ver_bytes (bytes)   : Key net version bytes

        Returns:
            str: Serialized key
        """

        # Serialize key
        ser_key = (
            key_net_ver_bytes
            + bytes(key_data.Depth()) + bytes(key_data.ParentFingerPrint()) + bytes(key_data.Index())
            + bytes(key_data.ChainCode()) + key_bytes
        )
        # Encode it
        return Base58Encoder.CheckEncode(ser_key)


class Bip32PrivateKeySerializer:
    """
    BIP32 private key serializer class.
    It serializes private keys.
    """

    @staticmethod
    def Serialize(priv_key: IPrivateKey,
                  key_data: Bip32KeyData,
                  key_net_ver: Bip32KeyNetVersions = Bip32Const.MAIN_NET_KEY_NET_VERSIONS) -> str:
        """
        Serialize a private key.

        Args:
            priv_key (IPrivateKey object)                     : IPrivateKey object
            key_data (BipKeyData object)                      : Key data
            key_net_ver (Bip32KeyNetVersions object, optional): Key net versions (BIP32 main net version by default)

        Returns:
            str: Serialized private key
        """
        return _Bip32KeySerializer.Serialize(b"\x00" + priv_key.Raw().ToBytes(),
                                             key_data,
                                             key_net_ver.Private())


class Bip32PublicKeySerializer:
    """
    BIP32 public key serializer class.
    It serializes public keys.
    """

    @staticmethod
    def Serialize(pub_key: IPublicKey,
                  key_data: Bip32KeyData,
                  key_net_ver: Bip32KeyNetVersions = Bip32Const.MAIN_NET_KEY_NET_VERSIONS) -> str:
        """
        Serialize a public key.

        Args:
            pub_key (IPublicKey object)                       : IPublicKey object
            key_data (BipKeyData object)                      : Key data
            key_net_ver (Bip32KeyNetVersions object, optional): Key net versions (BIP32 main net version by default)

        Returns:
            str: Serialized public key
        """
        return _Bip32KeySerializer.Serialize(pub_key.RawCompressed().ToBytes(),
                                             key_data,
                                             key_net_ver.Public())


class Bip32DeserializedKey:
    """
    BIP32 deserialized key class.
    It represents a key deserialized with the Bip32KeyDeserializer.
    """

    m_key_bytes: bytes
    m_key_data: Bip32KeyData
    m_is_public: bool

    def __init__(self,
                 key_bytes: bytes,
                 key_data: Bip32KeyData,
                 is_public: bool) -> None:
        """
        Construct class.

        Args:
            key_bytes (bytes)           : Key bytes
            key_data (BipKeyData object): Key data
            is_public (bool)            : True if the key is public, false otherwise

        Returns:
            str: Serialized public key
        """
        self.m_key_bytes = key_bytes
        self.m_key_data = key_data
        self.m_is_public = is_public

    def KeyBytes(self) -> bytes:
        """
        Get key bytes.

        Returns:
            bytes: Key bytes
        """
        return self.m_key_bytes

    def KeyData(self) -> Bip32KeyData:
        """
        Get key data.

        Returns:
            Bip32KeyData object: Bip32KeyData object
        """
        return self.m_key_data

    def IsPublic(self) -> bool:
        """
        Get if public.

        Returns:
            bool: True if the key is public, false otherwise
        """
        return self.m_is_public


class Bip32KeyDeserializer:
    """
    BIP32 key deserializer class.
    It deserializes an extended key.
    """

    @classmethod
    def DeserializeKey(cls,
                       ser_key_str: str,
                       key_net_ver: Bip32KeyNetVersions = Bip32Const.MAIN_NET_KEY_NET_VERSIONS) -> Bip32DeserializedKey:
        """
        Deserialize a key.

        Args:
            ser_key_str (str)                                 : Serialized key string
            key_net_ver (Bip32KeyNetVersions object, optional): Key net versions (BIP32 main net version by default)

        Returns:
            Bip32DeserializedKey object: Bip32DeserializedKey object

        Raises:
            Bip32KeyError: If the key is not valid
        """

        # Decode key
        ser_key_bytes = Base58Decoder.CheckDecode(ser_key_str)

        # Get if key is public/private depending on the net version
        is_public = cls.__GetIfPublic(ser_key_bytes, key_net_ver)

        # Validate length
        if is_public and len(ser_key_bytes) != Bip32KeySerConst.SERIALIZED_PUB_KEY_BYTE_LEN:
            raise Bip32KeyError(f"Invalid extended public key (wrong length: {len(ser_key_bytes)})")
        if not is_public and len(ser_key_bytes) not in Bip32KeySerConst.SERIALIZED_PRIV_KEY_BYTE_LEN:
            raise Bip32KeyError(f"Invalid extended private key (wrong length: {len(ser_key_bytes)})")

        # Get parts back
        key_bytes, key_data = cls.__GetPartsFromBytes(ser_key_bytes, is_public)

        return Bip32DeserializedKey(key_bytes, key_data, is_public)

    @staticmethod
    def __GetIfPublic(ser_key_bytes: bytes,
                      key_net_ver: Bip32KeyNetVersions) -> bool:
        """
        Get if the key is public.

        Args:
            ser_key_bytes (bytes)                   : Serialized key bytes
            key_net_ver (Bip32KeyNetVersions object): Key net versions

        Returns:
            bool: True if public, false otherwise

        Raises:
            Bip32KeyError: If the key net version is not valid
        """
        key_net_ver_got = ser_key_bytes[:Bip32KeyNetVersions.Length()]
        if key_net_ver_got == key_net_ver.Public():
            is_public = True
        elif key_net_ver_got == key_net_ver.Private():
            is_public = False
        else:
            raise Bip32KeyError(
                f"Invalid extended key (wrong net version: {BytesUtils.ToHexString(key_net_ver_got)})"
            )
        return is_public

    @staticmethod
    def __GetPartsFromBytes(ser_key_bytes: bytes,
                            is_public: bool) -> Tuple[bytes, Bip32KeyData]:
        """
        Get back key parts from serialized key bytes.

        Args:
            ser_key_bytes (bytes): Serialized key bytes
            is_public (bool)     : True if the key is public, false otherwise

        Returns:
            tuple[bytes, Bip32KeyData]: key bytes (index 0) and key data (index 1)

        Raises:
            Bip32KeyError: If the private key first byte is not zero
        """

        # Compute indexes
        depth_idx = Bip32KeyNetVersions.Length()
        fprint_idx = depth_idx + Bip32Depth.FixedLength()
        key_index_idx = fprint_idx + Bip32FingerPrint.FixedLength()
        chain_code_idx = key_index_idx + Bip32KeyIndex.FixedLength()
        key_idx = chain_code_idx + Bip32ChainCode.FixedLength()

        # Get parts
        depth = ser_key_bytes[depth_idx]
        fprint_bytes = ser_key_bytes[fprint_idx:key_index_idx]
        key_index_bytes = ser_key_bytes[key_index_idx:chain_code_idx]
        chain_code_bytes = ser_key_bytes[chain_code_idx:key_idx]
        key_bytes = ser_key_bytes[key_idx:]
        key_data = Bip32KeyData(Bip32Depth(depth),
                                Bip32KeyIndex.FromBytes(key_index_bytes),
                                Bip32ChainCode(chain_code_bytes),
                                Bip32FingerPrint(fprint_bytes))

        # If private key, the first byte shall be zero and shall be removed
        if not is_public:
            if key_bytes[0] != 0:
                raise Bip32KeyError(f"Invalid extended private key (wrong secret: {key_bytes[0]})")
            key_bytes = key_bytes[1:]

        return key_bytes, key_data



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for SLIP-0044 coin types.
Not all the coin types are defined, but only the used ones.
Reference: https://github.com/satoshilabs/slips/blob/master/slip-0044.md
"""


class Slip44:
    """
    SLIP-0044 class.
    It defines the coin types in according to SLIP-0044.
    """

    BITCOIN: int = 0
    TESTNET: int = 1
    LITECOIN: int = 2
    DOGECOIN: int = 3
    DASH: int = 5
    ETHEREUM: int = 60
    ETHEREUM_CLASSIC: int = 61
    ICON: int = 74
    VERGE: int = 77
    ATOM: int = 118
    MONERO: int = 128
    ZCASH: int = 133
    RIPPLE: int = 144
    BITCOIN_CASH: int = 145
    STELLAR: int = 148
    NANO: int = 165
    EOS: int = 194
    TRON: int = 195
    BITCOIN_SV: int = 236
    NIMIQ: int = 242
    ALGORAND: int = 283
    ZILLIQA: int = 313
    TERRA: int = 330
    POLKADOT: int = 354
    NEAR_PROTOCOL: int = 397
    ERGO: int = 429
    KUSAMA: int = 434
    KAVA: int = 459
    FILECOIN: int = 461
    BAND_PROTOCOL: int = 494
    THETA: int = 500
    SOLANA: int = 501
    ELROND: int = 508
    SECRET_NETWORK: int = 529
    NINE_CHRONICLES: int = 567
    APTOS: int = 637
    BINANCE_CHAIN: int = 714
    SUI: int = 784
    VECHAIN: int = 818
    NEO: int = 888
    OKEX_CHAIN: int = 996
    HARMONY_ONE: int = 1023
    ONTOLOGY: int = 1024
    TEZOS: int = 1729
    CARDANO: int = 1815
    AVALANCHE: int = 9000
    CELO: int = 52752
    PI_NETWORK: int = 314159



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with some ECDSA keys constants."""


class EcdsaKeysConst:
    """Class container for ECDSA keys constants."""

    # Point coordinate length in bytes
    POINT_COORD_BYTE_LEN: int = 32
    # Private key length in bytes
    PRIV_KEY_BYTE_LEN: int = 32
    # Uncompressed public key prefix
    PUB_KEY_UNCOMPRESSED_PREFIX: bytes = b"\x04"
    # Compressed public key length in bytes
    PUB_KEY_COMPRESSED_BYTE_LEN: int = 33
    # Uncompressed public key length in bytes
    PUB_KEY_UNCOMPRESSED_BYTE_LEN: int = 65



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with helper class for data bytes."""

# Imports
from typing import Iterator

from .bytes import BytesUtils
from ..typing import Literal


class DataBytes:
    """
    Data bytes class.
    It allows to get bytes in different formats.
    """

    m_data_bytes: bytes

    def __init__(self,
                 data_bytes: bytes) -> None:
        """
        Construct class.

        Args:
            data_bytes (bytes): Data bytes
        """
        self.m_data_bytes = data_bytes

    def Length(self) -> int:
        """
        Get length in bytes.

        Returns:
            int: Length in bytes
        """
        return len(self.m_data_bytes)

    def Size(self) -> int:
        """
        Get length in bytes (same of Length()).

        Returns:
            int: Length in bytes
        """
        return self.Length()

    def ToBytes(self) -> bytes:
        """
        Get data bytes.

        Returns:
            bytes: Data bytes
        """
        return self.m_data_bytes

    def ToHex(self) -> str:
        """
        Get data bytes in hex format.

        Returns:
            str: Data bytes in hex format
        """
        return BytesUtils.ToHexString(self.m_data_bytes)

    def ToInt(self,
              endianness: Literal["little", "big"] = "big") -> int:
        """
        Get data bytes as an integer.

        Args:
            endianness ("big" or "little", optional): Endianness (default: big)

        Returns:
            int: Data bytes as an integer
        """
        return BytesUtils.ToInteger(self.m_data_bytes, endianness)

    def __len__(self) -> int:
        """
        Get length in bytes.

        Returns:
            int: Length in bytes
        """
        return self.Length()

    def __bytes__(self) -> bytes:
        """
        Get data bytes.

        Returns:
            bytes: Data bytes
        """
        return self.ToBytes()

    def __int__(self) -> int:
        """
        Get data bytes as integer.

        Returns:
            bytes: Data bytes as integer
        """
        return self.ToInt()

    def __repr__(self) -> str:
        """
        Get data bytes representation.

        Returns:
            str: Data bytes representation
        """
        return self.ToHex()

    def __getitem__(self,
                    idx: int) -> int:
        """
        Get the element with the specified index.

        Args:
            idx (int): Index

        Returns:
            int: Element

        Raises:
            IndexError: If the index is not valid
        """
        return self.m_data_bytes[idx]

    def __iter__(self) -> Iterator[int]:
        """
        Get the iterator to the current element.

        Returns:
            Iterator object: Iterator to the current element
        """
        yield from self.m_data_bytes

    def __eq__(self,
               other: object) -> bool:
        """
        Equality operator.

        Args:
            other (bytes, str, int or DataBytes object): Other object to compare

        Returns:
            bool: True if equal false otherwise

        Raises:
            TypeError: If the other object is not of the correct type
        """
        if not isinstance(other, (bytes, int, str, DataBytes)):
            raise TypeError(f"Invalid type for checking equality ({type(other)})")

        if isinstance(other, bytes):
            return other == bytes(self)
        if isinstance(other, int):
            return other == int(self)
        if isinstance(other, str):
            return other == str(self)
        return bytes(other) == bytes(self)



# from .bip_bitcoin_cash_conf import BipBitcoinCashConf
# from .bip_coin_conf import BipCoinConf, BipCoinFctCallsConf
from .bip_coin_conf import BipCoinConf
from .bip_coins import BipCoins
from .bip_conf_const import (
    DER_PATH_HARDENED_FULL, DER_PATH_HARDENED_MID, DER_PATH_HARDENED_SHORT, DER_PATH_NON_HARDENED_FULL
)
# from .bip_litecoin_conf import BipLitecoinConf



from __future__ import annotations

from dataclasses import dataclass, fields, make_dataclass
from typing import Dict, Optional, Tuple


@dataclass(frozen=True, eq=False)
class TupleDataclass:
    """
    Dataclass that behaves like a tuple at the same time. Used when data has defined order and names.
    For instance in case of named tuples or function responses.
    """

    # getattr is called when attribute is not found in object. For instance when using object.unknown_attribute.
    # This way pyright will know that there might be some arguments it doesn't know about and will stop complaining
    # about some fields that don't exist statically.
    def __getattr__(self, item):
        # This should always fail - only attributes that don't exist end up in here.
        # We use __getattribute__ to get the native error.
        return super().__getattribute__(item)

    def __getitem__(self, item: int):
        field = fields(self)[item]
        return getattr(self, field.name)

    def __iter__(self):
        return (getattr(self, field.name) for field in fields(self))

    def as_tuple(self) -> Tuple:
        """
        Creates a regular tuple from TupleDataclass.
        """
        return tuple(self)

    def as_dict(self) -> Dict:
        """
        Creates a regular dict from TupleDataclass.
        """
        return {field.name: getattr(self, field.name) for field in fields(self)}

    # Added for backward compatibility with previous implementation based on NamedTuple
    def _asdict(self):
        return self.as_dict()

    def __eq__(self, other):
        if isinstance(other, TupleDataclass):
            return self.as_tuple() == other.as_tuple()
        return self.as_tuple() == other

    @staticmethod
    def from_dict(data: Dict, *, name: Optional[str] = None) -> TupleDataclass:
        result_class = make_dataclass(
            name or "TupleDataclass",
            fields=[(key, type(value)) for key, value in data.items()],
            bases=(TupleDataclass,),
            frozen=True,
            eq=False,
        )
        return result_class(**data)



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for BIP32 Khovratovich/Law keys derivation.
Reference: https://github.com/LedgerHQ/orakolo/blob/master/papers/Ed25519_BIP%20Final.pdf
"""

# Imports
from .bip32_ex import Bip32KeyError
from .bip32_key_data import Bip32KeyIndex
from .bip32_keys import Bip32PublicKey
from .kholaw.bip32_kholaw_key_derivator_base import Bip32KholawEd25519KeyDerivatorBase
from bip_utils.ecc import Ed25519KholawPrivateKey, EllipticCurve, IPoint
from ..utils.misc import BytesUtils, IntegerUtils


class Bip32KholawEd25519KeyDerivator(Bip32KholawEd25519KeyDerivatorBase):
    """
    BIP32 Khovratovich/Law ed25519 key derivator class.
    It allows keys derivation for ed25519 curves in according to BIP32 Khovratovich/Law.
    """

    @staticmethod
    def _SerializeIndex(index: Bip32KeyIndex) -> bytes:
        """
        Serialize key index.

        Args:
            index (Bip32KeyIndex object): Key index

        Returns:
            bytes: Serialized index
        """
        return index.ToBytes(endianness="little")

    @staticmethod
    def _NewPrivateKeyLeftPart(zl_bytes: bytes,
                               kl_bytes: bytes,
                               curve: EllipticCurve) -> bytes:
        """
        Compute the new private key left part for private derivation.

        Args:
            zl_bytes (bytes)            : Leftmost Z 32-byte
            kl_bytes (bytes)            : Leftmost private key 32-byte
            curve (EllipticCurve object): EllipticCurve object

        Returns:
            bytes: Leftmost new private key 32-byte
        """
        zl_int = BytesUtils.ToInteger(zl_bytes[:28], endianness="little")
        kl_int = BytesUtils.ToInteger(kl_bytes, endianness="little")

        prvl_int = (zl_int * 8) + kl_int
        # Discard child if multiple of curve order
        if prvl_int % curve.Order() == 0:
            raise Bip32KeyError("Computed child key is not valid, very unlucky index")

        return IntegerUtils.ToBytes(prvl_int,
                                    bytes_num=Ed25519KholawPrivateKey.Length() // 2,
                                    endianness="little")

    @staticmethod
    def _NewPrivateKeyRightPart(zr_bytes: bytes,
                                kr_bytes: bytes) -> bytes:
        """
        Compute the new private key right part for private derivation.

        Args:
            zr_bytes (bytes): Rightmost Z 32-byte
            kr_bytes (bytes): Rightmost private key 32-byte

        Returns:
            bytes: Rightmost new private key 32-byte
        """
        zr_int = BytesUtils.ToInteger(zr_bytes, endianness="little")
        kpr_int = BytesUtils.ToInteger(kr_bytes, endianness="little")
        kr_int = (zr_int + kpr_int) % (2 ** 256)

        return IntegerUtils.ToBytes(kr_int,
                                    bytes_num=Ed25519KholawPrivateKey.Length() // 2,
                                    endianness="little")

    @staticmethod
    def _NewPublicKeyPoint(pub_key: Bip32PublicKey,
                           zl_bytes: bytes) -> IPoint:
        """
        Compute new public key point for public derivation.

        Args:
            pub_key (Bip32PublicKey object): Bip32PublicKey object
            zl_bytes (bytes)               : Leftmost Z 32-byte

        Returns:
            IPoint object: IPoint object
        """

        # Compute the new public key point: PKEY + 8ZL * G
        zl_int = BytesUtils.ToInteger(zl_bytes[:28], endianness="little")
        return pub_key.Point() + ((zl_int * 8) * pub_key.Curve().Generator())



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with helper class for Base32."""

# Imports
import base64
import binascii
from typing import Optional, Union

from .algo import AlgoUtils


class Base32Const:
    """Class container for Base32 constants."""

    # Alphabet
    ALPHABET: str = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567"
    # Padding character
    PADDING_CHAR: str = "="


class _Base32Utils:
    """
    Base32 utility class.
    It provides some helper methods for decoding/encoding Base32 format.
    """

    @staticmethod
    def AddPadding(data: str) -> str:
        """
        Add padding to an encoded Base32 string.
        Used if the string was encoded with Base32Encoder.EncodeNoPadding

        Args:
            data (str): Data

        Returns:
            str: Padded string
        """
        last_block_width = len(data) % 8
        if last_block_width != 0:
            data += (8 - last_block_width) * Base32Const.PADDING_CHAR
        return data

    @staticmethod
    def TranslateAlphabet(data: str,
                          from_alphabet: str,
                          to_alphabet: str) -> str:
        """
        Translate the standard Base32 alphabet to a custom one.

        Args:
            data (str)         : Data
            from_alphabet (str): Starting alphabet string
            to_alphabet (str)  : Final alphabet string

        Returns:
            str: String with translated alphabet
        """
        return data.translate(str.maketrans(from_alphabet, to_alphabet))


class Base32Decoder:
    """
    Base32 decoder class.
    It provides methods for decoding to Base32 format.
    """

    @staticmethod
    def Decode(data: str,
               custom_alphabet: Optional[str] = None) -> bytes:
        """
        Decode from Base32.

        Args:
            data (str)                     : Data
            custom_alphabet (str, optional): Custom alphabet string

        Returns:
            bytes: Decoded bytes

        Raises:
            ValueError: If the Base32 string is not valid
        """
        try:
            data_dec = _Base32Utils.AddPadding(data)
            if custom_alphabet is not None:
                data_dec = _Base32Utils.TranslateAlphabet(data_dec, custom_alphabet, Base32Const.ALPHABET)

            return base64.b32decode(data_dec)
        except binascii.Error as ex:
            raise ValueError("Invalid Base32 string") from ex


class Base32Encoder:
    """
    Base32 encoder class.
    It provides methods for encoding to Base32 format.
    """

    @staticmethod
    def Encode(data: Union[bytes, str],
               custom_alphabet: Optional[str] = None) -> str:
        """
        Encode to Base32.

        Args:
            data (str or bytes)            : Data
            custom_alphabet (str, optional): Custom alphabet string

        Returns:
            str: Encoded string
        """
        b32_enc = AlgoUtils.Decode(base64.b32encode(AlgoUtils.Encode(data)))
        if custom_alphabet is not None:
            b32_enc = _Base32Utils.TranslateAlphabet(b32_enc, Base32Const.ALPHABET, custom_alphabet)

        return b32_enc

    @staticmethod
    def EncodeNoPadding(data: Union[bytes, str],
                        custom_alphabet: Optional[str] = None) -> str:
        """
        Encode to Base32 by removing the final padding.

        Args:
            data (str or bytes)            : Data
            custom_alphabet (str, optional): Custom alphabet string

        Returns:
            str: Encoded string
        """
        return Base32Encoder.Encode(data, custom_alphabet).rstrip(Base32Const.PADDING_CHAR)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for BIP32 SLIP-0010 master key generation.
Reference: https://github.com/satoshilabs/slips/blob/master/slip-0010.md
"""

# Imports
from typing import Tuple

from ..base import IBip32MstKeyGenerator
from ...ecc import EllipticCurveGetter, EllipticCurveTypes
from ...utils.crypto import HmacSha512


class Bip32Slip10MstKeyGeneratorConst:
    """Class container for BIP32 SLIP-0010 master key generator constants."""

    # Minimum length in bytes for seed
    SEED_MIN_BYTE_LEN: int = 16
    # HMAC keys for different curves
    HMAC_KEY_ED25519_BYTES: bytes = b"ed25519 seed"
    HMAC_KEY_NIST256P1_BYTES: bytes = b"Nist256p1 seed"
    HMAC_KEY_SECP256K1_BYTES: bytes = b"Bitcoin seed"


class _Bip32Slip10MstKeyGenerator:
    """
    BIP32 SLIP-0010 generic master key generator class.
    It allows master keys generation in according to BIP32 SLIP-0010.
    """

    @staticmethod
    def GenerateFromSeed(seed_bytes: bytes,
                         hmac_key_bytes: bytes,
                         curve_type: EllipticCurveTypes) -> Tuple[bytes, bytes]:
        """
        Generate a master key from the specified seed and return a Bip32Base object.

        Args:
            seed_bytes (bytes)                      : Seed bytes
            hmac_key_bytes (bytes)                  : HMAC key bytes

        Returns:
            Bip32Base object: Bip32Base object

        Raises:
            Bip32KeyError: If the seed is not suitable for master key generation
            ValueError: If seed length is not valid
        """
        if len(seed_bytes) < Bip32Slip10MstKeyGeneratorConst.SEED_MIN_BYTE_LEN:
            raise ValueError(f"Invalid seed length ({len(seed_bytes)})")

        hmac_half_len = HmacSha512.DigestSize() // 2
        priv_key_cls = EllipticCurveGetter.FromType(curve_type).PrivateKeyClass()

        # Compute HMAC, retry if the resulting private key is not valid
        hmac = b""
        hmac_data = seed_bytes
        success = False

        while not success:
            hmac = HmacSha512.QuickDigest(hmac_key_bytes, hmac_data)
            # If private key is not valid, the new HMAC data is the current HMAC
            success = priv_key_cls.IsValidBytes(hmac[:hmac_half_len])
            if not success:
                hmac_data = hmac

        return hmac[:hmac_half_len], hmac[hmac_half_len:]


class Bip32Slip10Ed2519MstKeyGenerator(IBip32MstKeyGenerator):
    """
    BIP32 SLIP-0010 ed25519 master key generator class.
    It allows master keys generation in according to BIP32 SLIP-0010 for ed25519 curve.
    """

    @classmethod
    def GenerateFromSeed(cls,
                         seed_bytes: bytes) -> Tuple[bytes, bytes]:
        """
        Generate a master key from the specified seed.

        Args:
            seed_bytes (bytes): Seed bytes

        Returns:
            tuple[bytes, bytes]: Private key bytes (index 0) and chain code bytes (index 1)

        Raises:
            Bip32KeyError: If the seed is not suitable for master key generation
            ValueError: If seed length is not valid
        """
        return _Bip32Slip10MstKeyGenerator.GenerateFromSeed(seed_bytes,
                                                            Bip32Slip10MstKeyGeneratorConst.HMAC_KEY_ED25519_BYTES,
                                                            EllipticCurveTypes.ED25519)


class Bip32Slip10Nist256p1MstKeyGenerator(IBip32MstKeyGenerator):
    """
    BIP32 SLIP-0010 nist256p1 master key generator class.
    It allows master keys generation in according to BIP32 SLIP-0010 for nist256p1 curve.
    """

    @classmethod
    def GenerateFromSeed(cls,
                         seed_bytes: bytes) -> Tuple[bytes, bytes]:
        """
        Generate a master key from the specified seed.

        Args:
            seed_bytes (bytes): Seed bytes

        Returns:
            tuple[bytes, bytes]: Private key bytes (index 0) and chain code bytes (index 1)

        Raises:
            Bip32KeyError: If the seed is not suitable for master key generation
            ValueError: If seed length is not valid
        """
        return _Bip32Slip10MstKeyGenerator.GenerateFromSeed(seed_bytes,
                                                            Bip32Slip10MstKeyGeneratorConst.HMAC_KEY_NIST256P1_BYTES,
                                                            EllipticCurveTypes.NIST256P1)


class Bip32Slip10Secp256k1MstKeyGenerator(IBip32MstKeyGenerator):
    """
    BIP32 SLIP-0010 secp256k1 master key generator class.
    It allows master keys generation in according to BIP32 SLIP-0010 for secp256k1 curve.
    """

    @classmethod
    def GenerateFromSeed(cls,
                         seed_bytes: bytes) -> Tuple[bytes, bytes]:
        """
        Generate a master key from the specified seed.

        Args:
            seed_bytes (bytes): Seed bytes

        Returns:
            tuple[bytes, bytes]: Private key bytes (index 0) and chain code bytes (index 1)

        Raises:
            Bip32KeyError: If the seed is not suitable for master key generation
            ValueError: If seed length is not valid
        """
        return _Bip32Slip10MstKeyGenerator.GenerateFromSeed(seed_bytes,
                                                            Bip32Slip10MstKeyGeneratorConst.HMAC_KEY_SECP256K1_BYTES,
                                                            EllipticCurveTypes.SECP256K1)



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for SHA-2 algorithms."""

# Imports
import hashlib
from typing import Any, Union

from Crypto.Hash import SHA512

from ..misc import AlgoUtils


HASHLIB_USE_SHA512_256: bool = "sha512_256" in hashlib.algorithms_available


class Sha256:
    """
    SHA256 class.
    It computes digests using SHA256 algorithm.
    """

    handle: Any

    def __init__(self) -> None:
        """Construct class."""
        self.handle = hashlib.sha256()

    def Update(self,
               data_bytes: bytes) -> None:
        """
        Update digest.

        Args:
            data_bytes (bytes): Data bytes
        """
        self.handle.update(data_bytes)

    def Digest(self) -> bytes:
        """
        Get the computed digest.

        Returns:
            bytes: Computed digest
        """
        return self.handle.digest()

    @staticmethod
    def QuickDigest(data: Union[bytes, str]) -> bytes:
        """
        Compute the digest (quick version).

        Args:
            data (str or bytes): Data

        Returns:
            bytes: Computed digest
        """
        return hashlib.sha256(AlgoUtils.Encode(data)).digest()

    @staticmethod
    def DigestSize() -> int:
        """
        Get the digest size in bytes.

        Returns:
            int: Digest size in bytes
        """
        return hashlib.sha256().digest_size


class DoubleSha256:
    """
    Double SHA256 class.
    It computes digests using SHA256 algorithm twice.
    """

    @staticmethod
    def QuickDigest(data: Union[bytes, str]) -> bytes:
        """
        Compute the digest (quick version).

        Args:
            data (str or bytes): Data

        Returns:
            bytes: Computed digest
        """
        return Sha256.QuickDigest(Sha256.QuickDigest(data))

    @staticmethod
    def DigestSize() -> int:
        """
        Get the digest size in bytes.

        Returns:
            int: Digest size in bytes
        """
        return Sha256.DigestSize()


class Sha512:
    """
    SHA512 class.
    It computes digests using SHA512 algorithm.
    """

    @staticmethod
    def QuickDigest(data: Union[bytes, str]) -> bytes:
        """
        Compute the digest (quick version).

        Args:
            data (str or bytes): Data

        Returns:
            bytes: Computed digest
        """
        return hashlib.sha512(AlgoUtils.Encode(data)).digest()

    @staticmethod
    def DigestSize() -> int:
        """
        Get the digest size in bytes.

        Returns:
            int: Digest size in bytes
        """
        return hashlib.sha512().digest_size


class Sha512_256:   # noqa: N801
    """
    SHA512/256 class.
    It computes digests using SHA512/256 algorithm.
    """

    @staticmethod
    def QuickDigest(data: Union[bytes, str]) -> bytes:
        """
        Compute the digest (quick version).

        Args:
            data (str or bytes): Data

        Returns:
            bytes: Computed digest
        """
        if HASHLIB_USE_SHA512_256:
            return hashlib.new("sha512_256", AlgoUtils.Encode(data)).digest()
        # Use Cryptodome if not implemented in hashlib
        return SHA512.new(AlgoUtils.Encode(data), truncate="256").digest()

    @staticmethod
    def DigestSize() -> int:
        """
        Get the digest size in bytes.

        Returns:
            int: Digest size in bytes
        """
        return (hashlib.new("sha512_256").digest_size
                if HASHLIB_USE_SHA512_256
                else SHA512.new(truncate="256").digest_size)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with generic coins configuration for all other modules."""

# Imports
from .coin_conf import CoinConf
from ..slip.slip173 import Slip173
from ..utils.conf import CoinNames


# Bitcoin constants used by different coins
# Main net
_BTC_P2PKH_NET_VER_MN: bytes = b"\x00"
_BTC_P2SH_NET_VER_MN: bytes = b"\x05"
_BTC_P2WPKH_HRP_MN: str = Slip173.BITCOIN_MAINNET
_BTC_P2WPKH_WIT_VER_MN: int = 0
_BTC_P2TR_HRP_MN: str = Slip173.BITCOIN_MAINNET
_BTC_P2TR_WIT_VER_MN: int = 1
_BTC_WIF_NET_VER_MN: bytes = b"\x80"
# Test net
_BTC_P2PKH_NET_VER_TN: bytes = b"\x6f"
_BTC_P2SH_NET_VER_TN: bytes = b"\xc4"
_BTC_P2WPKH_HRP_TN: str = Slip173.BITCOIN_TESTNET
_BTC_P2WPKH_WIT_VER_TN: int = 0
_BTC_P2TR_HRP_TN: str = Slip173.BITCOIN_TESTNET
_BTC_P2TR_WIT_VER_TN: int = 1
_BTC_WIF_NET_VER_TN: bytes = b"\xef"
# Regtest
_BTC_P2PKH_NET_VER_RT: bytes = _BTC_P2PKH_NET_VER_TN
_BTC_P2SH_NET_VER_RT: bytes = _BTC_P2SH_NET_VER_TN
_BTC_P2WPKH_HRP_RT: str = Slip173.BITCOIN_REGTEST
_BTC_P2WPKH_WIT_VER_RT: int = _BTC_P2TR_WIT_VER_TN
_BTC_P2TR_HRP_RT: str = Slip173.BITCOIN_REGTEST
_BTC_P2TR_WIT_VER_RT: int = _BTC_P2TR_WIT_VER_TN
_BTC_WIF_NET_VER_RT: bytes = _BTC_WIF_NET_VER_TN


class CoinsConf:
    """Class container for coins configuration."""

    # # Configuration for Acala
    # Acala: CoinConf = CoinConf(
    #     coin_name=CoinNames("Acala", "ACA"),
    #     params={
    #         "addr_ss58_format": 10,
    #     },
    # )

    # # Configuration for Akash Network
    # AkashNetwork: CoinConf = CoinConf(
    #     coin_name=CoinNames("Akash Network", "AKT"),
    #     params={
    #         "addr_hrp": Slip173.AKASH_NETWORK,
    #     },
    # )

    # # Configuration for Algorand
    # Algorand: CoinConf = CoinConf(
    #     coin_name=CoinNames("Algorand", "ALGO"),
    #     params={},
    # )

    # # Configuration for Aptos
    # Aptos: CoinConf = CoinConf(
    #     coin_name=CoinNames("Aptos", "APTOS"),
    #     params={
    #         "addr_prefix": "0x",
    #     },
    # )

    # # Configuration for Arbitrum
    # Arbitrum: CoinConf = CoinConf(
    #     coin_name=CoinNames("Arbitrum", "ARB"),
    #     params={},
    # )

    # # Configuration for Avax C-Chain
    # AvaxCChain: CoinConf = CoinConf(
    #     coin_name=CoinNames("Avax C-Chain", "AVAX"),
    #     params={},
    # )

    # # Configuration for Avax P-Chain
    # AvaxPChain: CoinConf = CoinConf(
    #     coin_name=CoinNames("Avax P-Chain", "AVAX"),
    #     params={
    #         "addr_hrp": "avax",
    #         "addr_prefix": "P-",
    #     },
    # )

    # # Configuration for Avax X-Chain
    # AvaxXChain: CoinConf = CoinConf(
    #     coin_name=CoinNames("Avax X-Chain", "AVAX"),
    #     params={
    #         "addr_hrp": "avax",
    #         "addr_prefix": "X-",
    #     },
    # )

    # # Configuration for Axelar
    # Axelar: CoinConf = CoinConf(
    #     coin_name=CoinNames("Axelar", "AXL"),
    #     params={
    #         "addr_hrp": Slip173.AXELAR,
    #     },
    # )

    # # Configuration for Band Protocol
    # BandProtocol: CoinConf = CoinConf(
    #     coin_name=CoinNames("Band Protocol", "BAND"),
    #     params={
    #         "addr_hrp": Slip173.BAND_PROTOCOL,
    #     },
    # )

    # # Configuration for Bifrost
    # Bifrost: CoinConf = CoinConf(
    #     coin_name=CoinNames("Bifrost", "BNC"),
    #     params={
    #         "addr_ss58_format": 6,
    #     },
    # )

    # # Configuration for Binance Chain
    # BinanceChain: CoinConf = CoinConf(
    #     coin_name=CoinNames("Binance Chain", "BNB"),
    #     params={
    #         "addr_hrp": Slip173.BINANCE_CHAIN,
    #     },
    # )

    # # Configuration for Binance Smart Chain
    # BinanceSmartChain: CoinConf = CoinConf(
    #     coin_name=CoinNames("Binance Smart Chain", "BNB"),
    #     params={},
    # )

    # Configuration for Bitcoin main net
    BitcoinMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Bitcoin", "BTC"),
        params={
            "p2pkh_net_ver": _BTC_P2PKH_NET_VER_MN,
            "p2sh_net_ver": _BTC_P2SH_NET_VER_MN,
            "p2wpkh_hrp": _BTC_P2WPKH_HRP_MN,
            "p2wpkh_wit_ver": _BTC_P2WPKH_WIT_VER_MN,
            "p2tr_hrp": _BTC_P2TR_HRP_MN,
            "p2tr_wit_ver": _BTC_P2TR_WIT_VER_MN,
            "wif_net_ver": _BTC_WIF_NET_VER_MN,
        },
    )

    # # Configuration for Bitcoin test net
    # BitcoinTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Bitcoin TestNet", "BTC"),
    #     params={
    #         "p2pkh_net_ver": _BTC_P2PKH_NET_VER_TN,
    #         "p2sh_net_ver": _BTC_P2SH_NET_VER_TN,
    #         "p2wpkh_hrp": _BTC_P2WPKH_HRP_TN,
    #         "p2wpkh_wit_ver": _BTC_P2WPKH_WIT_VER_TN,
    #         "p2tr_hrp": _BTC_P2TR_HRP_TN,
    #         "p2tr_wit_ver": _BTC_P2TR_WIT_VER_TN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_TN,
    #     },
    # )

    # # Configuration for Bitcoin regtest
    # BitcoinRegTest: CoinConf = CoinConf(
    #     coin_name=CoinNames("Bitcoin RegTest", "BTC"),
    #     params={
    #         "p2pkh_net_ver": _BTC_P2PKH_NET_VER_RT,
    #         "p2sh_net_ver": _BTC_P2SH_NET_VER_RT,
    #         "p2wpkh_hrp": _BTC_P2WPKH_HRP_RT,
    #         "p2wpkh_wit_ver": _BTC_P2WPKH_WIT_VER_RT,
    #         "p2tr_hrp": _BTC_P2TR_HRP_RT,
    #         "p2tr_wit_ver": _BTC_P2TR_WIT_VER_RT,
    #         "wif_net_ver": _BTC_WIF_NET_VER_RT,
    #     },
    # )

    # # Configuration for Bitcoin Cash main net
    # BitcoinCashMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Bitcoin Cash", "BCH"),
    #     params={
    #         "p2pkh_std_hrp": "bitcoincash",
    #         "p2pkh_std_net_ver": _BTC_P2PKH_NET_VER_MN,
    #         "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_MN,
    #         "p2sh_std_hrp": "bitcoincash",
    #         "p2sh_std_net_ver": b"\x08",
    #         "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_MN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_MN,
    #     },
    # )

    # # Configuration for Bitcoin Cash test net
    # BitcoinCashTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Bitcoin Cash TestNet", "BCH"),
    #     params={
    #         "p2pkh_std_hrp": "bchtest",
    #         "p2pkh_std_net_ver": b"\x00",
    #         "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_TN,
    #         "p2sh_std_hrp": "bchtest",
    #         "p2sh_std_net_ver": b"\x08",
    #         "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_TN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_TN,
    #     },
    # )

    # # Configuration for Bitcoin Cash Simple Ledger Protocol main net
    # BitcoinCashSlpMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Bitcoin Cash SLP", "SLP"),
    #     params={
    #         "p2pkh_std_hrp": "simpleledger",
    #         "p2pkh_std_net_ver": b"\x00",
    #         "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_MN,
    #         "p2sh_std_hrp": "simpleledger",
    #         "p2sh_std_net_ver": b"\x08",
    #         "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_MN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_MN,
    #     },
    # )

    # # Configuration for Bitcoin Cash Simple Ledger Protocol test net
    # BitcoinCashSlpTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Bitcoin Cash SLP TestNet", "SLP"),
    #     params={
    #         "p2pkh_std_hrp": "slptest",
    #         "p2pkh_std_net_ver": b"\x00",
    #         "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_TN,
    #         "p2sh_std_hrp": "slptest",
    #         "p2sh_std_net_ver": b"\x08",
    #         "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_TN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_TN,
    #     },
    # )

    # # Configuration for Bitcoin SV main net
    # BitcoinSvMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("BitcoinSV", "BSV"),
    #     params={
    #         "p2pkh_net_ver": _BTC_P2PKH_NET_VER_MN,
    #         "p2sh_net_ver": _BTC_P2SH_NET_VER_MN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_MN,
    #     },
    # )

    # # Configuration for Bitcoin SV test net
    # BitcoinSvTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("BitcoinSV TestNet", "BSV"),
    #     params={
    #         "p2pkh_net_ver": _BTC_P2PKH_NET_VER_TN,
    #         "p2sh_net_ver": _BTC_P2SH_NET_VER_TN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_TN,
    #     },
    # )

    # # Configuration for Cardano main net
    # CardanoMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Cardano", "ADA"),
    #     params={
    #         "addr_hrp": "addr",
    #         "staking_addr_hrp": "stake",
    #     },
    # )

    # # Configuration for Cardano test
    # CardanoTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Cardano TestNet", "ADA"),
    #     params={
    #         "addr_hrp": "addr_test",
    #         "staking_addr_hrp": "stake_test",
    #     },
    # )

    # # Configuration for Celestia
    # Celestia: CoinConf = CoinConf(
    #     coin_name=CoinNames("Celestia", "TIA"),
    #     params={
    #         "addr_hrp": Slip173.CELESTIA,
    #     },
    # )

    # # Configuration for Celo
    # Celo: CoinConf = CoinConf(
    #     coin_name=CoinNames("Celo", "CELO"),
    #     params={},
    # )

    # # Configuration for Certik
    # Certik: CoinConf = CoinConf(
    #     coin_name=CoinNames("Certik", "CTK"),
    #     params={
    #         "addr_hrp": Slip173.CERTIK,
    #     },
    # )

    # # Configuration for ChainX
    # ChainX: CoinConf = CoinConf(
    #     coin_name=CoinNames("ChainX", "PCX"),
    #     params={
    #         "addr_ss58_format": 44,
    #     },
    # )

    # # Configuration for Chihuahua
    # Chihuahua: CoinConf = CoinConf(
    #     coin_name=CoinNames("Chihuahua", "HUAHUA"),
    #     params={
    #         "addr_hrp": Slip173.CHIHUAHUA,
    #     },
    # )

    # Configuration for Cosmos
    Cosmos: CoinConf = CoinConf(
        coin_name=CoinNames("Cosmos", "ATOM"),
        params={
            "addr_hrp": Slip173.COSMOS,
        },
    )

    # # Configuration for Dash main net
    # DashMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Dash", "DASH"),
    #     params={
    #         "p2pkh_net_ver": b"\x4c",
    #         "p2sh_net_ver": b"\x10",
    #         "wif_net_ver": b"\xcc",
    #     },
    # )

    # # Configuration for Dash test net
    # DashTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Dash TestNet", "DASH"),
    #     params={
    #         "p2pkh_net_ver": b"\x8c",
    #         "p2sh_net_ver": b"\x13",
    #         "wif_net_ver": _BTC_WIF_NET_VER_TN,
    #     },
    # )

    # # Configuration for Dogecoin main net
    # DogecoinMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Dogecoin", "DOGE"),
    #     params={
    #         "p2pkh_net_ver": b"\x1e",
    #         "p2sh_net_ver": b"\x16",
    #         "wif_net_ver": b"\x9e",
    #     },
    # )

    # # Configuration for Dogecoin test net
    # DogecoinTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Dogecoin TestNet", "DOGE"),
    #     params={
    #         "p2pkh_net_ver": b"\x71",
    #         "p2sh_net_ver": _BTC_P2SH_NET_VER_TN,
    #         "wif_net_ver": b"\xf1",
    #     },
    # )

    # # Configuration for dYdX
    # DYDX: CoinConf = CoinConf(
    #     coin_name=CoinNames("dYdX", "DYDX"),
    #     params={
    #         "addr_hrp": Slip173.DYDX,
    #     },
    # )

    # # Configuration for eCash main net
    # EcashMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("eCash", "XEC"),
    #     params={
    #         "p2pkh_std_hrp": "ecash",
    #         "p2pkh_std_net_ver": b"\x00",
    #         "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_MN,
    #         "p2sh_std_hrp": "ecash",
    #         "p2sh_std_net_ver": b"\x08",
    #         "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_MN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_MN,
    #     },
    # )

    # # Configuration for eCash test net
    # EcashTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("eCash TestNet", "XEC"),
    #     params={
    #         "p2pkh_std_hrp": "ectest",
    #         "p2pkh_std_net_ver": b"\x00",
    #         "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_TN,
    #         "p2sh_std_hrp": "ectest",
    #         "p2sh_std_net_ver": b"\x08",
    #         "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_TN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_TN,
    #     },
    # )

    # # Configuration for Edgeware
    # Edgeware: CoinConf = CoinConf(
    #     coin_name=CoinNames("Edgeware", "EDG"),
    #     params={
    #         "addr_ss58_format": 7,
    #     },
    # )

    # # Configuration for Elrond
    # Elrond: CoinConf = CoinConf(
    #     coin_name=CoinNames("MultiversX", "EGLD"),
    #     params={
    #         "addr_hrp": Slip173.ELROND,
    #     },
    # )

    # # Configuration for Eos
    # Eos: CoinConf = CoinConf(
    #     coin_name=CoinNames("EOS", "EOS"),
    #     params={
    #         "addr_prefix": "EOS",
    #     },
    # )

    # # Configuration for Ergo main net
    # ErgoMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Ergo", "ERGO"),
    #     params={},
    # )

    # # Configuration for Ergo test net
    # ErgoTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Ergo TestNet", "ERGO"),
    #     params={},
    # )

    # # Configuration for Ethereum
    # Ethereum: CoinConf = CoinConf(
    #     coin_name=CoinNames("Ethereum", "ETH"),
    #     params={
    #         "addr_prefix": "0x",
    #     },
    # )

    # # Configuration for Ethereum Classic
    # EthereumClassic: CoinConf = CoinConf(
    #     coin_name=CoinNames("Ethereum Classic", "ETC"),
    #     params={},
    # )

    # # Configuration for Fantom Opera
    # FantomOpera: CoinConf = CoinConf(
    #     coin_name=CoinNames("Fantom Opera", "FTM"),
    #     params={},
    # )

    # # Configuration for Fetch.ai
    # FetchAi: CoinConf = CoinConf(
    #     coin_name=CoinNames("Fetch.ai", "FET"),
    #     params={
    #         "addr_hrp": Slip173.FETCH_AI,
    #     },
    # )

    # # Configuration for Filecoin
    # Filecoin: CoinConf = CoinConf(
    #     coin_name=CoinNames("Filecoin", "FIL"),
    #     params={
    #         "addr_prefix": "f",
    #     },
    # )

    # # Configuration for generic Substrate coin
    # GenericSubstrate: CoinConf = CoinConf(
    #     coin_name=CoinNames("Generic Substrate", ""),
    #     params={
    #         "addr_ss58_format": 42,
    #     },
    # )

    # # Configuration for Harmony One
    # HarmonyOne: CoinConf = CoinConf(
    #     coin_name=CoinNames("Harmony One", "ONE"),
    #     params={
    #         "addr_hrp": Slip173.HARMONY_ONE,
    #     },
    # )

    # # Configuration for Huobi Chain
    # HuobiChain: CoinConf = CoinConf(
    #     coin_name=CoinNames("Huobi Token", "HT"),
    #     params={},
    # )

    # # Configuration for Icon
    # Icon: CoinConf = CoinConf(
    #     coin_name=CoinNames("Icon", "ICX"),
    #     params={
    #         "addr_prefix": "hx",
    #     },
    # )

    # # Configuration for Injective
    # Injective: CoinConf = CoinConf(
    #     coin_name=CoinNames("Injective", "INJ"),
    #     params={
    #         "addr_hrp": Slip173.INJECTIVE,
    #     },
    # )

    # # Configuration for IRISnet
    # IrisNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("IRIS Network", "IRIS"),
    #     params={
    #         "addr_hrp": Slip173.IRIS_NETWORK,
    #     },
    # )

    # # Configuration for Karura
    # Karura: CoinConf = CoinConf(
    #     coin_name=CoinNames("Karura", "KAR"),
    #     params={
    #         "addr_ss58_format": 8,
    #     },
    # )

    # # Configuration for Kava
    # Kava: CoinConf = CoinConf(
    #     coin_name=CoinNames("Kava", "KAVA"),
    #     params={
    #         "addr_hrp": Slip173.KAVA,
    #     },
    # )

    # # Configuration for Kusama
    # Kusama: CoinConf = CoinConf(
    #     coin_name=CoinNames("Kusama", "KSM"),
    #     params={
    #         "addr_ss58_format": 2,
    #     },
    # )

    # # Configuration for Litecoin main net
    # LitecoinMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Litecoin", "LTC"),
    #     params={
    #         "p2pkh_std_net_ver": b"\x30",
    #         "p2pkh_depr_net_ver": _BTC_P2PKH_NET_VER_MN,
    #         "p2sh_std_net_ver": b"\x32",
    #         "p2sh_depr_net_ver": _BTC_P2SH_NET_VER_MN,
    #         "p2wpkh_hrp": Slip173.LITECOIN_MAINNET,
    #         "p2wpkh_wit_ver": _BTC_P2WPKH_WIT_VER_MN,
    #         "wif_net_ver": b"\xb0",
    #     },
    # )

    # # Configuration for Litecoin test net
    # LitecoinTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Litecoin TestNet", "LTC"),
    #     params={
    #         "p2pkh_std_net_ver": b"\x6f",
    #         "p2pkh_depr_net_ver": _BTC_P2PKH_NET_VER_TN,
    #         "p2sh_std_net_ver": b"\x3a",
    #         "p2sh_depr_net_ver": _BTC_P2SH_NET_VER_TN,
    #         "p2wpkh_hrp": Slip173.LITECOIN_TESTNET,
    #         "p2wpkh_wit_ver": _BTC_P2WPKH_WIT_VER_TN,
    #         "wif_net_ver": _BTC_WIF_NET_VER_TN,
    #     },
    # )

    # # Configuration for Metis
    # Metis: CoinConf = CoinConf(
    #     coin_name=CoinNames("Metis", "METIS"),
    #     params={},
    # )

    # # Configuration for Monero main net
    # MoneroMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Monero", "XMR"),
    #     params={
    #         "addr_net_ver": b"\x12",
    #         "addr_int_net_ver": b"\x13",
    #         "subaddr_net_ver": b"\x2a",
    #     },
    # )

    # # Configuration for Monero stage net
    # MoneroStageNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Monero StageNet", "XMR"),
    #     params={
    #         "addr_net_ver": b"\x18",
    #         "addr_int_net_ver": b"\x19",
    #         "subaddr_net_ver": b"\x24",
    #     },
    # )

    # # Configuration for Monero test net
    # MoneroTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Monero TestNet", "XMR"),
    #     params={
    #         "addr_net_ver": b"\x35",
    #         "addr_int_net_ver": b"\x36",
    #         "subaddr_net_ver": b"\x3f",
    #     },
    # )

    # # Configuration for Moonbeam
    # Moonbeam: CoinConf = CoinConf(
    #     coin_name=CoinNames("Moonbeam", "GLMR"),
    #     params={
    #         "addr_ss58_format": 1284,
    #     },
    # )

    # # Configuration for Moonriver
    # Moonriver: CoinConf = CoinConf(
    #     coin_name=CoinNames("Moonriver", "MOVR"),
    #     params={
    #         "addr_ss58_format": 1285,
    #     },
    # )

    # # Configuration for Nano
    # Nano: CoinConf = CoinConf(
    #     coin_name=CoinNames("Nano", "NANO"),
    #     params={
    #         "addr_prefix": "nano_",
    #     },
    # )

    # # Configuration for Near Protocol
    # NearProtocol: CoinConf = CoinConf(
    #     coin_name=CoinNames("Near Protocol", "NEAR"),
    #     params={},
    # )

    # # For compatibility, later assigned to NeoLegacy
    # Neo: CoinConf

    # # Configuration for Neo legacy
    # NeoLegacy: CoinConf = CoinConf(
    #     coin_name=CoinNames("NEO", "NEO"),
    #     params={
    #         "addr_ver": b"\x17",
    #         "addr_prefix": b"\x21",
    #         "addr_suffix": b"\xac",
    #         "wif_net_ver": _BTC_WIF_NET_VER_MN,
    #     },
    # )

    # # Configuration for Neo N3
    # NeoN3: CoinConf = CoinConf(
    #     coin_name=CoinNames("NEO", "NEO"),
    #     params={
    #         "addr_ver": b"\x35",
    #         "addr_prefix": b"\x0c\x21",
    #         "addr_suffix": b"\x41\x56\xe7\xb3\x27",
    #         "wif_net_ver": _BTC_WIF_NET_VER_MN,
    #     },
    # )

    # # Configuration for Neutron
    # Neutron: CoinConf = CoinConf(
    #     coin_name=CoinNames("Neutron", "NTRN"),
    #     params={
    #         "addr_hrp": Slip173.NEUTRON,
    #     },
    # )

    # # Configuration for Nimiq
    # Nimiq: CoinConf = CoinConf(
    #     coin_name=CoinNames("Nimiq", "NIM"),
    #     params={
    #         "addr_prefix": "NQ"
    #     },
    # )

    # # Configuration for Nine Chronicles Gold
    # NineChroniclesGold: CoinConf = CoinConf(
    #     coin_name=CoinNames("NineChroniclesGold", "NCG"),
    #     params={},
    # )

    # # Configuration for OKEx Chain
    # OkexChain: CoinConf = CoinConf(
    #     coin_name=CoinNames("OKExChain", "OKT"),
    #     params={
    #         "addr_hrp": Slip173.OKEX_CHAIN,
    #     },
    # )

    # # Configuration for Ontology
    # Ontology: CoinConf = CoinConf(
    #     coin_name=CoinNames("Ontology", "ONT"),
    #     params={
    #         "addr_ver": b"\x17",
    #     },
    # )

    # # Configuration for Optimism
    # Optimism: CoinConf = CoinConf(
    #     coin_name=CoinNames("Optimism", "OP"),
    #     params={},
    # )

    # # Configuration for Osmosis
    # Osmosis: CoinConf = CoinConf(
    #     coin_name=CoinNames("Osmosis", "OSMO"),
    #     params={
    #         "addr_hrp": Slip173.OSMOSIS,
    #     },
    # )

    # # Configuration for Phala
    # Phala: CoinConf = CoinConf(
    #     coin_name=CoinNames("Phala Network", "PHA"),
    #     params={
    #         "addr_ss58_format": 30,
    #     },
    # )

    # # Configuration for Pi Network
    # PiNetwork: CoinConf = CoinConf(
    #     coin_name=CoinNames("Pi Network", "PI"),
    #     params={},
    # )

    # # Configuration for Plasm
    # Plasm: CoinConf = CoinConf(
    #     coin_name=CoinNames("Plasm Network", "PLM"),
    #     params={
    #         "addr_ss58_format": 5,
    #     },
    # )

    # # Configuration for Polkadot
    # Polkadot: CoinConf = CoinConf(
    #     coin_name=CoinNames("Polkadot", "DOT"),
    #     params={
    #         "addr_ss58_format": 0,
    #     },
    # )

    # # Configuration for Polygon
    # Polygon: CoinConf = CoinConf(
    #     coin_name=CoinNames("Polygon", "MATIC"),
    #     params={},
    # )

    # # Configuration for Ripple
    # Ripple: CoinConf = CoinConf(
    #     coin_name=CoinNames("Ripple", "XRP"),
    #     params={
    #         "p2pkh_net_ver": _BTC_P2PKH_NET_VER_MN,
    #     },
    # )

    # # Configuration for Secret Network
    # SecretNetwork: CoinConf = CoinConf(
    #     coin_name=CoinNames("Secret Network", "SCRT"),
    #     params={
    #         "addr_hrp": Slip173.SECRET_NETWORK,
    #     },
    # )

    # # Configuration for Solana
    # Solana: CoinConf = CoinConf(
    #     coin_name=CoinNames("Solana", "SOL"),
    #     params={},
    # )

    # # Configuration for Sora
    # Sora: CoinConf = CoinConf(
    #     coin_name=CoinNames("Sora", "XOR"),
    #     params={
    #         "addr_ss58_format": 69,
    #     },
    # )

    # # Configuration for Stafi
    # Stafi: CoinConf = CoinConf(
    #     coin_name=CoinNames("Stafi", "FIS"),
    #     params={
    #         "addr_hrp": Slip173.STAFI,
    #         "addr_ss58_format": 20,
    #     },
    # )

    # # Configuration for Stellar
    # Stellar: CoinConf = CoinConf(
    #     coin_name=CoinNames("Stellar", "XLM"),
    #     params={},
    # )

    # # Configuration for Sui
    # Sui: CoinConf = CoinConf(
    #     coin_name=CoinNames("Sui", "SUI"),
    #     params={
    #         "addr_prefix": "0x",
    #     },
    # )

    # # Configuration for Terra
    # Terra: CoinConf = CoinConf(
    #     coin_name=CoinNames("Terra", "LUNA"),
    #     params={
    #         "addr_hrp": Slip173.TERRA,
    #     },
    # )

    # # Configuration for Tezos
    # Tezos: CoinConf = CoinConf(
    #     coin_name=CoinNames("Tezos", "XTZ"),
    #     params={},
    # )

    # # Configuration for Theta
    # Theta: CoinConf = CoinConf(
    #     coin_name=CoinNames("Theta Network", "THETA"),
    #     params={},
    # )

    # # Configuration for Tron
    # Tron: CoinConf = CoinConf(
    #     coin_name=CoinNames("Tron", "TRX"),
    #     params={
    #         "addr_prefix": b"\x41",
    #     },
    # )

    # # Configuration for VeChain
    # VeChain: CoinConf = CoinConf(
    #     coin_name=CoinNames("VeChain", "VET"),
    #     params={},
    # )

    # # Configuration for Verge
    # Verge: CoinConf = CoinConf(
    #     coin_name=CoinNames("Verge", "XVG"),
    #     params={
    #         "p2pkh_net_ver": b"\x1e",
    #         "wif_net_ver": b"\x9e",
    #     },
    # )

    # # Configuration for Zcash main net
    # ZcashMainNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Zcash", "ZEC"),
    #     params={
    #         "p2pkh_net_ver": b"\x1c\xb8",
    #         "p2sh_net_ver": b"\x1c\xbd",
    #         "wif_net_ver": _BTC_WIF_NET_VER_MN,
    #     },
    # )

    # # Configuration for Zcash test net
    # ZcashTestNet: CoinConf = CoinConf(
    #     coin_name=CoinNames("Zcash TestNet", "ZEC"),
    #     params={
    #         "p2pkh_net_ver": b"\x1d\x25",
    #         "p2sh_net_ver": b"\x1c\xba",
    #         "wif_net_ver": _BTC_WIF_NET_VER_TN,
    #     },
    # )

    # # Configuration for Zilliqa
    # Zilliqa: CoinConf = CoinConf(
    #     coin_name=CoinNames("Zilliqa", "ZIL"),
    #     params={
    #         "addr_hrp": Slip173.ZILLIQA,
    #     },
    # )


# For compatibility
# CoinsConf.Neo = CoinsConf.NeoLegacy



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with interface for address encoding classes."""

# Imports
from abc import ABC, abstractmethod
from typing import Any


class IAddrDecoder(ABC):
    """Address decoder interface."""

    @staticmethod
    @abstractmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode an address to bytes.
        Depending on the coin, the result can be a public key or a public key hash bytes.

        Args:
            addr (str): Address string
            **kwargs  : Arbitrary arguments depending on the address type

        Returns:
            bytes: Public key bytes or public key hash

        Raises:
            ValueError: If the address encoding is not valid
        """
